﻿using System;
using Microsoft.Win32.SafeHandles;
using System.Runtime.InteropServices;
using System.IO;
using System.Windows.Forms;

public class DeviceInfo
{
    public override string ToString()
    {
        return name + " (unit " + getUnitNo() + ")";
    }
    public int getUnitNo()
    {
        return ((productID & 0x0F) + 1);
    }
    public DeviceInfo(SafeHandle fp, string path, string name, short vendorID, short productID, short version)
    {
        // remember the settings
        this.fp = fp;
        this.path = path;
        this.name = name;
        this.vendorID = vendorID;
        this.productID = productID;
        this.version = version;
        this.plungerEnabled = true;

        // read a status report
        byte[] buf = ReadUSB();
        if (buf != null)
        {
            // parse the reponse
            this.plungerEnabled = (buf[1] & 0x01) != 0;
        }
    }

    public DeviceInfo(DeviceInfo d)
    {
        this.path = d.path;
        this.name = d.name;
        this.vendorID = d.vendorID;
        this.productID = d.productID;
        this.version = d.version;
        this.plungerEnabled = d.plungerEnabled;
        this.fp = OpenFile();
    }

    private System.Threading.NativeOverlapped ov;
    public byte[] ReadUSB()
    {
        for (int tries = 0 ; tries < 3 ; ++tries)
        {
            const int rptLen = 15;
            byte[] buf = new byte[rptLen];
            buf[0] = 0x00;
            uint actual;
            if (HIDImports.ReadFile(fp.DangerousGetHandle(), buf, rptLen, out actual, ref ov) == 0)
            {
                // if the error is 6 ("invalid handle"), try re-opening the device
                if (TryReopenHandle())
                    continue;

                MessageBox.Show("Error reading from device: " + GetLastWin32ErrMsg());
                return null;
            }
            else if (actual != rptLen)
            {
                MessageBox.Show("Error reading from device: not all bytes received");
                return null;
            }
            else
                return buf;
        }

        // don't retry more than a few times
        return null;
    }


    private SafeHandle OpenFile()
    {
        return HIDImports.CreateFile(
            path, HIDImports.GENERIC_READ_WRITE, HIDImports.SHARE_READ_WRITE,
            IntPtr.Zero, FileMode.Open, 0, IntPtr.Zero);
    }

    private bool TryReopenHandle()
    {
        // if the last error is 6 ("invalid handle"), try re-opening it
        if (Marshal.GetLastWin32Error() == 6)
        {
            // try opening a new handle on the device path
            Console.WriteLine("invalid handle on read - trying to reopen handle");
            SafeHandle fp2 = OpenFile();

            // if that succeeded, replace the old handle with the new one and retry the read
            if (fp2 != null)
            {
                // replace the handle
                fp = fp2;

                // tell the caller to try again
                return true;
            }
        }

        // we didn't successfully reopen the handle
        return false;
    }
    
    public String GetLastWin32ErrMsg()
    {
        int errno = Marshal.GetLastWin32Error();
        return String.Format("{0} (Win32 error {1})", 
            new System.ComponentModel.Win32Exception(errno).Message, errno);
    }

    public bool SpecialRequest(byte id)
    {
        byte[] buf = new byte[9];
        buf[0] = 0x00;  // report ID - always 0
        buf[1] = 0x41;  // 0x41 -> Pinscape special request
        buf[2] = id;    // special request type
        return WriteUSB(buf);
    }

    public bool WriteUSB(byte[] buf)
    {
        for (int tries = 0; tries < 3; ++tries)
        {
            UInt32 actual;
            if (HIDImports.WriteFile(fp.DangerousGetHandle(), buf, 9, out actual, ref ov) == 0)
            {
                // try re-opening the handle, if it's an "invalid handle" error
                if (TryReopenHandle())
                    continue;

                MessageBox.Show("Error sending request to device: " + GetLastWin32ErrMsg());
                return false;
            }
            else if (actual != 9)
            {
                MessageBox.Show("Error sending request: not all bytes sent");
                return false;
            }
            else
                return true;
        }

        // maximum retries exceeded - return failure
        return false;
    }

    public SafeHandle fp;
    public string path;
    public string name;
    public short vendorID;
    public short productID;
    public short version;
    public bool plungerEnabled;
}
